@@ -2,7 +2,7 @@
  *
  *  Connection Manager
  *
- *  Copyright (C) 2007-2009  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2007-2010  Intel Corporation. All rights reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -86,8 +86,6 @@ static void connect_reply(DBusPendingCall *call, void *user_data)
 	DBG("network %p", network);
 
 	reply = dbus_pending_call_steal_reply(call);
-	if (reply == NULL)
-		return;
 
 	dbus_error_init(&error);
 
@@ -110,10 +108,15 @@ static void connect_reply(DBusPendingCall *call, void *user_data)
 	index = connman_inet_ifindex(interface);
 
 	connman_network_set_index(network, index);
+
+	connman_network_set_method(network, CONNMAN_IPCONFIG_METHOD_DHCP);
+
 	connman_network_set_connected(network, TRUE);
 
 done:
 	dbus_message_unref(reply);
+
+	dbus_pending_call_unref(call);
 }
 
 static int pan_connect(struct connman_network *network)
@@ -170,8 +173,6 @@ static void disconnect_reply(DBusPendingCall *call, void *user_data)
 	DBG("network %p", network);
 
 	reply = dbus_pending_call_steal_reply(call);
-	if (reply == NULL)
-		return;
 
 	dbus_error_init(&error);
 
@@ -189,6 +190,8 @@ static void disconnect_reply(DBusPendingCall *call, void *user_data)
 
 done:
 	dbus_message_unref(reply);
+
+	dbus_pending_call_unref(call);
 }
 
 static int pan_disconnect(struct connman_network *network)
@@ -243,99 +246,6 @@ static struct connman_network_driver pan_driver = {
 	.disconnect	= pan_disconnect,
 };
 
-static int bluetooth_probe(struct connman_device *device)
-{
-	DBG("device %p", device);
-
-	return 0;
-}
-
-static void bluetooth_remove(struct connman_device *device)
-{
-	DBG("device %p", device);
-}
-
-static void powered_reply(DBusPendingCall *call, void *user_data)
-{
-	DBusMessage *reply;
-
-	DBG("");
-
-	reply = dbus_pending_call_steal_reply(call);
-
-	dbus_message_unref(reply);
-}
-
-static int change_powered(DBusConnection *connection, const char *path,
-							dbus_bool_t powered)
-{
-	DBusMessage *message;
-	DBusMessageIter iter;
-	DBusPendingCall *call;
-
-	DBG("");
-
-	if (path == NULL)
-		return -EINVAL;
-
-	message = dbus_message_new_method_call(BLUEZ_SERVICE, path,
-					BLUEZ_ADAPTER_INTERFACE, SET_PROPERTY);
-	if (message == NULL)
-		return -ENOMEM;
-
-	dbus_message_set_auto_start(message, FALSE);
-
-	dbus_message_iter_init_append(message, &iter);
-	connman_dbus_property_append_variant(&iter, "Powered",
-						DBUS_TYPE_BOOLEAN, &powered);
-
-	if (dbus_connection_send_with_reply(connection, message,
-						&call, TIMEOUT) == FALSE) {
-		connman_error("Failed to change Powered property");
-		dbus_message_unref(message);
-		return -EINVAL;
-	}
-
-	if (call == NULL) {
-		connman_error("D-Bus connection not available");
-		dbus_message_unref(message);
-		return -EINVAL;
-	}
-
-	dbus_pending_call_set_notify(call, powered_reply, NULL, NULL);
-
-	dbus_message_unref(message);
-
-	return -EINPROGRESS;
-}
-
-static int bluetooth_enable(struct connman_device *device)
-{
-	const char *path = connman_device_get_string(device, "Path");
-
-	DBG("device %p", device);
-
-	return change_powered(connection, path, TRUE);
-}
-
-static int bluetooth_disable(struct connman_device *device)
-{
-	const char *path = connman_device_get_string(device, "Path");
-
-	DBG("device %p", device);
-
-	return change_powered(connection, path, FALSE);
-}
-
-static struct connman_device_driver bluetooth_driver = {
-	.name		= "bluetooth",
-	.type		= CONNMAN_DEVICE_TYPE_BLUETOOTH,
-	.probe		= bluetooth_probe,
-	.remove		= bluetooth_remove,
-	.enable		= bluetooth_enable,
-	.disable	= bluetooth_disable,
-};
-
 static void extract_properties(DBusMessage *reply, const char **parent,
 						const char **address,
 						const char **name,
@@ -430,8 +340,6 @@ static void network_properties_reply(DBusPendingCall *call, void *user_data)
 	char ident[13];
 
 	reply = dbus_pending_call_steal_reply(call);
-	if (reply == NULL)
-		return;
 
 	extract_properties(reply, &parent, &address, NULL, &name,
 						NULL, NULL, &uuids, NULL);
@@ -480,6 +388,8 @@ static void network_properties_reply(DBusPendingCall *call, void *user_data)
 
 done:
 	dbus_message_unref(reply);
+
+	dbus_pending_call_unref(call);
 }
 
 static void add_network(struct connman_device *device, const char *path)
@@ -535,7 +445,8 @@ static void check_networks(struct connman_device *device,
 	}
 }
 
-static void adapter_changed(DBusConnection *connection, DBusMessage *message)
+static gboolean adapter_changed(DBusConnection *connection,
+				DBusMessage *message, void *user_data)
 {
 	const char *path = dbus_message_get_path(message);
 	struct connman_device *device;
@@ -546,10 +457,10 @@ static void adapter_changed(DBusConnection *connection, DBusMessage *message)
 
 	device = g_hash_table_lookup(bluetooth_devices, path);
 	if (device == NULL)
-		return;
+		return TRUE;
 
 	if (dbus_message_iter_init(message, &iter) == FALSE)
-		return;
+		return TRUE;
 
 	dbus_message_iter_get_basic(&iter, &key);
 
@@ -569,6 +480,8 @@ static void adapter_changed(DBusConnection *connection, DBusMessage *message)
 	} else if (g_str_equal(key, "Devices") == TRUE) {
 		check_networks(device, &value);
 	}
+
+	return TRUE;
 }
 
 static void adapter_properties_reply(DBusPendingCall *call, void *user_data)
@@ -585,8 +498,6 @@ static void adapter_properties_reply(DBusPendingCall *call, void *user_data)
 	DBG("path %s", path);
 
 	reply = dbus_pending_call_steal_reply(call);
-	if (reply == NULL)
-		return;
 
 	if (path == NULL)
 		goto done;
@@ -619,6 +530,8 @@ static void adapter_properties_reply(DBusPendingCall *call, void *user_data)
 
 	connman_device_set_mode(device, CONNMAN_DEVICE_MODE_NETWORK_MULTIPLE);
 
+	connman_device_set_string(device, "Path", path);
+
 	if (connman_device_register(device) < 0) {
 		connman_device_unref(device);
 		goto done;
@@ -629,6 +542,7 @@ static void adapter_properties_reply(DBusPendingCall *call, void *user_data)
 update:
 	connman_device_set_string(device, "Address", address);
 	connman_device_set_string(device, "Name", name);
+	connman_device_set_string(device, "Path", path);
 
 	connman_device_set_powered(device, powered);
 	connman_device_set_scanning(device, scanning);
@@ -638,6 +552,8 @@ update:
 
 done:
 	dbus_message_unref(reply);
+
+	dbus_pending_call_unref(call);
 }
 
 static void add_adapter(DBusConnection *connection, const char *path)
@@ -672,6 +588,17 @@ done:
 	dbus_message_unref(message);
 }
 
+static gboolean adapter_added(DBusConnection *connection, DBusMessage *message,
+				void *user_data)
+{
+	const char *path;
+
+	dbus_message_get_args(message, NULL, DBUS_TYPE_OBJECT_PATH, &path,
+				DBUS_TYPE_INVALID);
+	add_adapter(connection, path);
+	return TRUE;
+}
+
 static void remove_adapter(DBusConnection *connection, const char *path)
 {
 	DBG("path %s", path);
@@ -679,6 +606,17 @@ static void remove_adapter(DBusConnection *connection, const char *path)
 	g_hash_table_remove(bluetooth_devices, path);
 }
 
+static gboolean adapter_removed(DBusConnection *connection, DBusMessage *message,
+				void *user_data)
+{
+	const char *path;
+
+	dbus_message_get_args(message, NULL, DBUS_TYPE_OBJECT_PATH, &path,
+				DBUS_TYPE_INVALID);
+	remove_adapter(connection, path);
+	return TRUE;
+}
+
 static void list_adapters_reply(DBusPendingCall *call, void *user_data)
 {
 	DBusMessage *reply;
@@ -689,8 +627,6 @@ static void list_adapters_reply(DBusPendingCall *call, void *user_data)
 	DBG("");
 
 	reply = dbus_pending_call_steal_reply(call);
-	if (reply == NULL)
-		return;
 
 	dbus_error_init(&error);
 
@@ -713,6 +649,8 @@ static void list_adapters_reply(DBusPendingCall *call, void *user_data)
 
 done:
 	dbus_message_unref(reply);
+
+	dbus_pending_call_unref(call);
 }
 
 static void unregister_device(gpointer data)
@@ -770,40 +708,108 @@ static void bluetooth_disconnect(DBusConnection *connection, void *user_data)
 	bluetooth_devices = NULL;
 }
 
-static DBusHandlerResult bluetooth_signal(DBusConnection *connection,
-					DBusMessage *message, void *user_data)
+static int bluetooth_probe(struct connman_device *device)
 {
-	if (dbus_message_is_signal(message, BLUEZ_ADAPTER_INTERFACE,
-						PROPERTY_CHANGED) == TRUE) {
-		adapter_changed(connection, message);
-	} else if (dbus_message_is_signal(message, BLUEZ_MANAGER_INTERFACE,
-						ADAPTER_ADDED) == TRUE) {
-		const char *path;
-		dbus_message_get_args(message, NULL,
-					DBUS_TYPE_OBJECT_PATH, &path,
-							DBUS_TYPE_INVALID);
-		add_adapter(connection, path);
-	} else if (dbus_message_is_signal(message, BLUEZ_MANAGER_INTERFACE,
-						ADAPTER_REMOVED) == TRUE) {
-		const char *path;
-		dbus_message_get_args(message, NULL,
-					DBUS_TYPE_OBJECT_PATH, &path,
-							DBUS_TYPE_INVALID);
-		remove_adapter(connection, path);
-	}
+	DBG("device %p", device);
 
-	return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
+	return 0;
 }
 
+static void bluetooth_remove(struct connman_device *device)
+{
+	DBG("device %p", device);
+}
+
+static void powered_reply(DBusPendingCall *call, void *user_data)
+{
+	DBusMessage *reply;
+
+	DBG("");
+
+	reply = dbus_pending_call_steal_reply(call);
+
+	dbus_message_unref(reply);
+
+	dbus_pending_call_unref(call);
+
+	add_adapter(connection, user_data);
+}
+
+static int change_powered(DBusConnection *connection, const char *path,
+							dbus_bool_t powered)
+{
+	DBusMessage *message;
+	DBusMessageIter iter;
+	DBusPendingCall *call;
+
+	DBG("");
+
+	if (path == NULL)
+		return -EINVAL;
+
+	message = dbus_message_new_method_call(BLUEZ_SERVICE, path,
+					BLUEZ_ADAPTER_INTERFACE, SET_PROPERTY);
+	if (message == NULL)
+		return -ENOMEM;
+
+	dbus_message_set_auto_start(message, FALSE);
+
+	dbus_message_iter_init_append(message, &iter);
+	connman_dbus_property_append_basic(&iter, "Powered",
+						DBUS_TYPE_BOOLEAN, &powered);
+
+	if (dbus_connection_send_with_reply(connection, message,
+						&call, TIMEOUT) == FALSE) {
+		connman_error("Failed to change Powered property");
+		dbus_message_unref(message);
+		return -EINVAL;
+	}
+
+	if (call == NULL) {
+		connman_error("D-Bus connection not available");
+		dbus_message_unref(message);
+		return -EINVAL;
+	}
+
+	dbus_pending_call_set_notify(call, powered_reply,
+					g_strdup(path), g_free);
+
+	dbus_message_unref(message);
+
+	return -EINPROGRESS;
+}
+
+static int bluetooth_enable(struct connman_device *device)
+{
+	const char *path = connman_device_get_string(device, "Path");
+
+	DBG("device %p", device);
+
+	return change_powered(connection, path, TRUE);
+}
+
+static int bluetooth_disable(struct connman_device *device)
+{
+	const char *path = connman_device_get_string(device, "Path");
+
+	DBG("device %p", device);
+
+	return change_powered(connection, path, FALSE);
+}
+
+static struct connman_device_driver bluetooth_driver = {
+	.name		= "bluetooth",
+	.type		= CONNMAN_DEVICE_TYPE_BLUETOOTH,
+	.probe		= bluetooth_probe,
+	.remove		= bluetooth_remove,
+	.enable		= bluetooth_enable,
+	.disable	= bluetooth_disable,
+};
+
 static guint watch;
-
-static const char *added_rule = "type=signal,member=" ADAPTER_ADDED
-					",interface=" BLUEZ_MANAGER_INTERFACE;
-static const char *removed_rule = "type=signal,member=" ADAPTER_REMOVED
-					",interface=" BLUEZ_MANAGER_INTERFACE;
-
-static const char *adapter_rule = "type=signal,member=" PROPERTY_CHANGED
-					",interface=" BLUEZ_ADAPTER_INTERFACE;
+static guint added_watch;
+static guint removed_watch;
+static guint adapter_watch;
 
 static int bluetooth_init(void)
 {
@@ -813,10 +819,28 @@ static int bluetooth_init(void)
 	if (connection == NULL)
 		return -EIO;
 
-	if (dbus_connection_add_filter(connection, bluetooth_signal,
-						NULL, NULL) == FALSE) {
+	watch = g_dbus_add_service_watch(connection, BLUEZ_SERVICE,
+			bluetooth_connect, bluetooth_disconnect, NULL, NULL);
+
+	added_watch = g_dbus_add_signal_watch(connection, NULL, NULL,
+						BLUEZ_MANAGER_INTERFACE,
+						ADAPTER_ADDED, adapter_added,
+						NULL, NULL);
+
+	removed_watch = g_dbus_add_signal_watch(connection, NULL, NULL,
+						BLUEZ_MANAGER_INTERFACE,
+						ADAPTER_REMOVED, adapter_removed,
+						NULL, NULL);
+
+	adapter_watch = g_dbus_add_signal_watch(connection, NULL, NULL,
+						BLUEZ_MANAGER_INTERFACE,
+						PROPERTY_CHANGED, adapter_changed,
+						NULL, NULL);
+
+	if (watch == 0 || added_watch == 0 || removed_watch == 0
+			|| adapter_watch == 0) {
 		err = -EIO;
-		goto unref;
+		goto remove;
 	}
 
 	err = connman_network_driver_register(&pan_driver);
@@ -829,26 +853,14 @@ static int bluetooth_init(void)
 		goto remove;
 	}
 
-	watch = g_dbus_add_service_watch(connection, BLUEZ_SERVICE,
-			bluetooth_connect, bluetooth_disconnect, NULL, NULL);
-	if (watch == 0) {
-		connman_device_driver_unregister(&bluetooth_driver);
-		connman_network_driver_unregister(&pan_driver);
-		err = -EIO;
-		goto remove;
-	}
-
-	dbus_bus_add_match(connection, added_rule, NULL);
-	dbus_bus_add_match(connection, removed_rule, NULL);
-	dbus_bus_add_match(connection, adapter_rule, NULL);
-	dbus_connection_flush(connection);
-
 	return 0;
 
 remove:
-	dbus_connection_remove_filter(connection, bluetooth_signal, NULL);
+	g_dbus_remove_watch(connection, watch);
+	g_dbus_remove_watch(connection, added_watch);
+	g_dbus_remove_watch(connection, removed_watch);
+	g_dbus_remove_watch(connection, adapter_watch);
 
-unref:
 	dbus_connection_unref(connection);
 
 	return err;
@@ -856,20 +868,16 @@ unref:
 
 static void bluetooth_exit(void)
 {
-	dbus_bus_remove_match(connection, adapter_rule, NULL);
-	dbus_bus_remove_match(connection, removed_rule, NULL);
-	dbus_bus_remove_match(connection, added_rule, NULL);
-	dbus_connection_flush(connection);
-
 	g_dbus_remove_watch(connection, watch);
+	g_dbus_remove_watch(connection, added_watch);
+	g_dbus_remove_watch(connection, removed_watch);
+	g_dbus_remove_watch(connection, adapter_watch);
 
 	bluetooth_disconnect(connection, NULL);
 
 	connman_device_driver_unregister(&bluetooth_driver);
 	connman_network_driver_unregister(&pan_driver);
 
-	dbus_connection_remove_filter(connection, bluetooth_signal, NULL);
-
 	dbus_connection_unref(connection);
 }
 
