@@ -2,7 +2,7 @@
  *
  *  Connection Manager
  *
- *  Copyright (C) 2007-2009  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2007-2010  Intel Corporation. All rights reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -162,6 +162,8 @@ struct supplicant_result {
 	dbus_uint16_t capabilities;
 	gboolean adhoc;
 	gboolean has_wep;
+	gboolean has_psk;
+	gboolean has_8021x;
 	gboolean has_wpa;
 	gboolean has_rsn;
 	gboolean has_wps;
@@ -175,7 +177,7 @@ struct supplicant_result {
 struct supplicant_task {
 	int ifindex;
 	char *ifname;
-	gboolean mac80211;
+	gboolean cfg80211;
 	struct connman_device *device;
 	struct connman_network *network;
 	struct connman_network *pending_network;
@@ -250,6 +252,15 @@ static int get_range(struct supplicant_task *task)
 
 	close(fd);
 
+	if (err < 0)
+		task->range->max_qual.updated |= IW_QUAL_ALL_INVALID;
+
+	connman_info("%s {scan} capabilities 0x%02x", task->ifname,
+						task->range->scan_capa);
+
+	connman_info("%s {quality} flags 0x%02x", task->ifname,
+					task->range->max_qual.updated);
+
 	return err;
 }
 
@@ -302,8 +313,6 @@ static void add_interface_reply(DBusPendingCall *call, void *user_data)
 	DBG("task %p", task);
 
 	reply = dbus_pending_call_steal_reply(call);
-	if (reply == NULL)
-		return;
 
 	if (dbus_message_get_type(reply) == DBUS_MESSAGE_TYPE_ERROR)
 		goto failed;
@@ -329,11 +338,15 @@ static void add_interface_reply(DBusPendingCall *call, void *user_data)
 
 	dbus_message_unref(reply);
 
+	dbus_pending_call_unref(call);
+
 	return;
 
 failed:
 	dbus_message_unref(reply);
 
+	dbus_pending_call_unref(call);
+
 	task_list = g_slist_remove(task_list, task);
 
 	connman_device_unref(task->device);
@@ -362,15 +375,12 @@ static int add_interface(struct supplicant_task *task)
 	dbus_message_iter_append_basic(&array,
 					DBUS_TYPE_STRING, &task->ifname);
 
-	dbus_message_iter_open_container(&array, DBUS_TYPE_ARRAY,
-			DBUS_DICT_ENTRY_BEGIN_CHAR_AS_STRING
-			DBUS_TYPE_STRING_AS_STRING DBUS_TYPE_VARIANT_AS_STRING
-			DBUS_DICT_ENTRY_END_CHAR_AS_STRING, &dict);
+	connman_dbus_dict_open(&array, &dict);
 
-	connman_dbus_dict_append_variant(&dict, "driver",
+	connman_dbus_dict_append_basic(&dict, "driver",
 						DBUS_TYPE_STRING, &driver);
 
-	dbus_message_iter_close_container(&array, &dict);
+	connman_dbus_dict_close(&array, &dict);
 
 	if (dbus_connection_send_with_reply(connection, message,
 						&call, TIMEOUT) == FALSE) {
@@ -402,8 +412,6 @@ static void get_interface_reply(DBusPendingCall *call, void *user_data)
 	DBG("task %p", task);
 
 	reply = dbus_pending_call_steal_reply(call);
-	if (reply == NULL)
-		return;
 
 	if (dbus_message_get_type(reply) == DBUS_MESSAGE_TYPE_ERROR) {
 		add_interface(task);
@@ -431,6 +439,8 @@ static void get_interface_reply(DBusPendingCall *call, void *user_data)
 
 done:
 	dbus_message_unref(reply);
+
+	dbus_pending_call_unref(call);
 }
 
 static int create_interface(struct supplicant_task *task)
@@ -488,6 +498,8 @@ static void remove_interface_reply(DBusPendingCall *call, void *user_data)
 	free_task(task);
 
 	dbus_message_unref(reply);
+
+	dbus_pending_call_unref(call);
 }
 
 static int remove_interface(struct supplicant_task *task)
@@ -749,6 +761,118 @@ static int disconnect_network(struct supplicant_task *task)
 	return 0;
 }
 
+static int set_network_tls(struct connman_network *network,
+			   DBusMessageIter *dict)
+{
+	const char *private_key, *client_cert, *ca_cert;
+	const char *private_key_password;
+
+	/*
+	 * For TLS, we at least need a key, the client cert,
+	 * and a passhprase.
+	 * Server cert is optional.
+	 */
+	client_cert = connman_network_get_string(network,
+						"WiFi.ClientCertFile");
+	if (client_cert == NULL)
+		return -EINVAL;
+
+	private_key = connman_network_get_string(network,
+						"WiFi.PrivateKeyFile");
+	if (private_key == NULL)
+		return -EINVAL;
+
+	private_key_password = connman_network_get_string(network,
+						"WiFi.PrivateKeyPassphrase");
+	if (private_key_password == NULL)
+		return -EINVAL;
+
+	ca_cert = connman_network_get_string(network, "WiFi.CACertFile");
+	if (ca_cert)
+		connman_dbus_dict_append_basic(dict, "ca_cert",
+						DBUS_TYPE_STRING, &ca_cert);
+
+	DBG("client cert %s private key %s", client_cert, private_key);
+
+	connman_dbus_dict_append_basic(dict, "private_key",
+						DBUS_TYPE_STRING, &private_key);
+	connman_dbus_dict_append_basic(dict, "private_key_passwd",
+							DBUS_TYPE_STRING,
+							&private_key_password);
+	connman_dbus_dict_append_basic(dict, "client_cert",
+						DBUS_TYPE_STRING, &client_cert);
+
+	return 0;
+}
+
+static int set_network_peap(struct connman_network *network,
+			    DBusMessageIter *dict, const char *passphrase)
+{
+	const char *client_cert, *ca_cert, *phase2;
+	char *phase2_auth;
+
+	/*
+	 * For PEAP, we at least need the sever cert, a 2nd
+	 * phase authentication and a passhprase.
+	 * Client cert is optional although strongly required
+	 * When setting the client cert, we then need a private
+	 * key as well.
+	 */
+	ca_cert = connman_network_get_string(network, "WiFi.CACertFile");
+	if (ca_cert == NULL)
+		return -EINVAL;
+
+	phase2 = connman_network_get_string(network, "WiFi.Phase2");
+	if (phase2 == NULL)
+		return -EINVAL;
+
+	DBG("CA cert %s phase2 auth %s", ca_cert, phase2);
+
+	client_cert = connman_network_get_string(network,
+							"WiFi.ClientCertFile");
+	if (client_cert) {
+		const char *private_key, *private_key_password;
+
+		private_key = connman_network_get_string(network,
+							"WiFi.PrivateKeyFile");
+		if (private_key == NULL)
+			return -EINVAL;
+
+		private_key_password =
+			connman_network_get_string(network,
+						"WiFi.PrivateKeyPassphrase");
+		if (private_key_password == NULL)
+			return -EINVAL;
+
+		connman_dbus_dict_append_basic(dict, "client_cert",
+						DBUS_TYPE_STRING, &client_cert);
+
+		connman_dbus_dict_append_basic(dict, "private_key",
+						DBUS_TYPE_STRING, &private_key);
+
+		connman_dbus_dict_append_basic(dict, "private_key_passwd",
+							DBUS_TYPE_STRING,
+							&private_key_password);
+
+		DBG("client cert %s private key %s", client_cert, private_key);
+	}
+
+	phase2_auth = g_strdup_printf("\"auth=%s\"", phase2);
+
+	connman_dbus_dict_append_basic(dict, "password",
+						DBUS_TYPE_STRING, &passphrase);
+
+	connman_dbus_dict_append_basic(dict, "ca_cert",
+						DBUS_TYPE_STRING, &ca_cert);
+
+	connman_dbus_dict_append_basic(dict, "phase2",
+						DBUS_TYPE_STRING, &phase2_auth);
+
+	g_free(phase2_auth);
+
+	return 0;
+}
+
 static int set_network(struct supplicant_task *task,
 				const unsigned char *network, int len,
 				const char *address, const char *security,
@@ -773,42 +897,104 @@ static int set_network(struct supplicant_task *task,
 
 	dbus_message_iter_init_append(message, &array);
 
-	dbus_message_iter_open_container(&array, DBUS_TYPE_ARRAY,
-			DBUS_DICT_ENTRY_BEGIN_CHAR_AS_STRING
-			DBUS_TYPE_STRING_AS_STRING DBUS_TYPE_VARIANT_AS_STRING
-			DBUS_DICT_ENTRY_END_CHAR_AS_STRING, &dict);
+	connman_dbus_dict_open(&array, &dict);
 
-	connman_dbus_dict_append_variant(&dict, "scan_ssid",
+	connman_dbus_dict_append_basic(&dict, "scan_ssid",
 					 DBUS_TYPE_UINT32, &scan_ssid);
 
-	if (address)
-		connman_dbus_dict_append_variant(&dict, "bssid",
+	if (network)
+		connman_dbus_dict_append_fixed_array(&dict, "ssid",
+						DBUS_TYPE_BYTE, &network, len);
+	else if (address)
+		connman_dbus_dict_append_basic(&dict, "bssid",
 						DBUS_TYPE_STRING, &address);
 
-	connman_dbus_dict_append_array(&dict, "ssid",
-					DBUS_TYPE_BYTE, &network, len);
-
-	if (g_ascii_strcasecmp(security, "wpa") == 0 ||
+	if (g_ascii_strcasecmp(security, "psk") == 0 ||
+				g_ascii_strcasecmp(security, "wpa") == 0 ||
 				g_ascii_strcasecmp(security, "rsn") == 0) {
 		const char *key_mgmt = "WPA-PSK";
-		connman_dbus_dict_append_variant(&dict, "key_mgmt",
+		connman_dbus_dict_append_basic(&dict, "key_mgmt",
 						DBUS_TYPE_STRING, &key_mgmt);
 
 		if (passphrase && strlen(passphrase) > 0)
-			connman_dbus_dict_append_variant(&dict, "psk",
+			connman_dbus_dict_append_basic(&dict, "psk",
 						DBUS_TYPE_STRING, &passphrase);
+	} else if (g_ascii_strcasecmp(security, "ieee8021x") == 0) {
+		struct connman_network *network = task->network;
+		const char *key_mgmt = "WPA-EAP", *eap, *identity;
+		char *eap_value;
+
+		/*
+		 * If our private key password is unset,
+		 * we use the supplied passphrase. That is needed
+		 * for PEAP where 2 passphrases (identity and client
+		 * cert may have to be provided.
+		 */
+		if (connman_network_get_string(network,
+					"WiFi.PrivateKeyPassphrase") == NULL)
+			connman_network_set_string(network,
+						"WiFi.PrivateKeyPassphrase",
+								passphrase);
+
+		eap = connman_network_get_string(network, "WiFi.EAP");
+		if (eap == NULL)
+			goto invalid;
+
+		/* We must have an identity for both PEAP and TLS */
+		identity = connman_network_get_string(network, "WiFi.Identity");
+		if (identity == NULL)
+			goto invalid;
+
+		DBG("key_mgmt %s eap %s identity %s", key_mgmt, eap, identity);
+
+		if (g_strcmp0(eap, "tls") == 0) {
+			int err;
+
+			err = set_network_tls(network, &dict);
+			if (err < 0) {
+				dbus_message_unref(message);
+				return err;
+			}
+		} else if (g_strcmp0(eap, "peap") == 0) {
+			int err;
+
+			err = set_network_peap(network, &dict, passphrase);
+			if (err < 0) {
+				dbus_message_unref(message);
+				return err;
+			}
+		} else {
+			connman_error("Unknown EAP %s", eap);
+			goto invalid;
+		}
+
+		/* wpa_supplicant only accepts upper case EAPs */
+		eap_value = g_ascii_strup(eap, -1);
+
+		connman_dbus_dict_append_basic(&dict, "key_mgmt",
+							DBUS_TYPE_STRING,
+							&key_mgmt);
+		connman_dbus_dict_append_basic(&dict, "eap",
+							DBUS_TYPE_STRING,
+							&eap_value);
+		connman_dbus_dict_append_basic(&dict, "identity",
+							DBUS_TYPE_STRING,
+							&identity);
+
+		g_free(eap_value);
+
 	} else if (g_ascii_strcasecmp(security, "wep") == 0) {
 		const char *key_mgmt = "NONE";
 		const char *auth_alg = "OPEN";
 		const char *key_index = "0";
 
-		if (task->mac80211 == TRUE)
+		if (task->cfg80211 == TRUE)
 			auth_alg = "OPEN SHARED";
 
-		connman_dbus_dict_append_variant(&dict, "auth_alg",
+		connman_dbus_dict_append_basic(&dict, "auth_alg",
 						DBUS_TYPE_STRING, &auth_alg);
 
-		connman_dbus_dict_append_variant(&dict, "key_mgmt",
+		connman_dbus_dict_append_basic(&dict, "key_mgmt",
 						DBUS_TYPE_STRING, &key_mgmt);
 
 		if (passphrase) {
@@ -825,25 +1011,25 @@ static int set_network(struct supplicant_task *task,
 					key[i] = (unsigned char) strtol(tmp,
 								NULL, 16);
 				}
-				connman_dbus_dict_append_array(&dict,
+				connman_dbus_dict_append_fixed_array(&dict,
 						"wep_key0", DBUS_TYPE_BYTE,
 							&key, size / 2);
 				free(key);
 			} else
-				connman_dbus_dict_append_variant(&dict,
+				connman_dbus_dict_append_basic(&dict,
 						"wep_key0", DBUS_TYPE_STRING,
 								&passphrase);
 
-			connman_dbus_dict_append_variant(&dict, "wep_tx_keyidx",
+			connman_dbus_dict_append_basic(&dict, "wep_tx_keyidx",
 						DBUS_TYPE_STRING, &key_index);
 		}
 	} else {
 		const char *key_mgmt = "NONE";
-		connman_dbus_dict_append_variant(&dict, "key_mgmt",
+		connman_dbus_dict_append_basic(&dict, "key_mgmt",
 						DBUS_TYPE_STRING, &key_mgmt);
 	}
 
-	dbus_message_iter_close_container(&array, &dict);
+	connman_dbus_dict_close(&array, &dict);
 
 	dbus_error_init(&error);
 
@@ -864,6 +1050,10 @@ static int set_network(struct supplicant_task *task,
 	dbus_message_unref(reply);
 
 	return 0;
+
+invalid:
+	dbus_message_unref(message);
+	return -EINVAL;
 }
 
 static void scan_reply(DBusPendingCall *call, void *user_data)
@@ -876,8 +1066,6 @@ static void scan_reply(DBusPendingCall *call, void *user_data)
 	task->scan_call = NULL;
 
 	reply = dbus_pending_call_steal_reply(call);
-	if (reply == NULL)
-		return;
 
 	if (dbus_message_get_type(reply) == DBUS_MESSAGE_TYPE_ERROR) {
 		connman_device_set_scanning(task->device, FALSE);
@@ -889,6 +1077,8 @@ static void scan_reply(DBusPendingCall *call, void *user_data)
 
 done:
 	dbus_message_unref(reply);
+
+	dbus_pending_call_unref(call);
 }
 
 
@@ -1066,6 +1256,76 @@ static void extract_ssid(DBusMessageIter *value,
 	}
 }
 
+static unsigned char wifi_oui[3]      = { 0x00, 0x50, 0xf2 };
+static unsigned char ieee80211_oui[3] = { 0x00, 0x0f, 0xac };
+
+static void extract_rsn(struct supplicant_result *result,
+					const unsigned char *buf, int len)
+{
+	uint16_t count;
+	int i;
+
+	/* Version */
+	if (len < 2)
+		return;
+
+	buf += 2;
+	len -= 2;
+
+	/* Group cipher */
+	if (len < 4)
+		return;
+
+	buf += 4;
+	len -= 4;
+
+	/* Pairwise cipher */
+	if (len < 2)
+		return;
+
+	count = buf[0] | (buf[1] << 8);
+	if (2 + (count * 4) > len)
+		return;
+
+	buf += 2 + (count * 4);
+	len -= 2 + (count * 4);
+
+	/* Authentication */
+	if (len < 2)
+		return;
+
+	count = buf[0] | (buf[1] << 8);
+	if (2 + (count * 4) > len)
+		return;
+
+	for (i = 0; i < count; i++) {
+		const unsigned char *ptr = buf + 2 + (i * 4);
+
+		if (memcmp(ptr, wifi_oui, 3) == 0) {
+			switch (ptr[3]) {
+			case 1:
+				result->has_8021x = TRUE;
+				break;
+			case 2:
+				result->has_psk = TRUE;
+				break;
+			}
+		} else if (memcmp(ptr, ieee80211_oui, 3) == 0) {
+			switch (ptr[3]) {
+			case 1:
+				result->has_8021x = TRUE;
+				break;
+			case 2:
+				result->has_psk = TRUE;
+				break;
+			}
+		}
+	}
+
+	buf += 2 + (count * 4);
+	len -= 2 + (count * 4);
+}
+
 static void extract_wpaie(DBusMessageIter *value,
 					struct supplicant_result *result)
 {
@@ -1076,8 +1336,10 @@ static void extract_wpaie(DBusMessageIter *value,
 	dbus_message_iter_recurse(value, &array);
 	dbus_message_iter_get_fixed_array(&array, &ie, &ie_len);
 
-	if (ie_len > 0)
+	if (ie_len > 6) {
 		result->has_wpa = TRUE;
+		extract_rsn(result, ie + 6, ie_len - 6);
+	}
 }
 
 static void extract_rsnie(DBusMessageIter *value,
@@ -1090,8 +1352,10 @@ static void extract_rsnie(DBusMessageIter *value,
 	dbus_message_iter_recurse(value, &array);
 	dbus_message_iter_get_fixed_array(&array, &ie, &ie_len);
 
-	if (ie_len > 0)
+	if (ie_len > 2) {
 		result->has_rsn = TRUE;
+		extract_rsn(result, ie + 2, ie_len - 2);
+	}
 }
 
 static void extract_wpsie(DBusMessageIter *value,
@@ -1125,7 +1389,7 @@ static void extract_capabilites(DBusMessageIter *value,
 static unsigned char calculate_strength(struct supplicant_task *task,
 					struct supplicant_result *result)
 {
-	if (task->range->max_qual.qual == 0) {
+	if (result->quality == -1 || task->range->max_qual.qual == 0) {
 		unsigned char strength;
 
 		if (result->level > 0)
@@ -1167,16 +1431,9 @@ static void properties_reply(DBusPendingCall *call, void *user_data)
 	DBG("task %p", task);
 
 	reply = dbus_pending_call_steal_reply(call);
-	if (reply == NULL) {
-		get_properties(task);
-		return;
-	}
 
-	if (dbus_message_get_type(reply) == DBUS_MESSAGE_TYPE_ERROR) {
-		dbus_message_unref(reply);
-		get_properties(task);
-		return;
-	}
+	if (dbus_message_get_type(reply) == DBUS_MESSAGE_TYPE_ERROR)
+		goto unref;
 
 	memset(&result, 0, sizeof(result));
 	result.frequency = -1;
@@ -1242,6 +1499,12 @@ static void properties_reply(DBusPendingCall *call, void *user_data)
 		dbus_message_iter_next(&dict);
 	}
 
+	DBG("capabilties %u frequency %d "
+			"quality %d noise %d level %d maxrate %d",
+					result.capabilities, result.frequency,
+						result.quality, result.noise,
+						result.level, result.maxrate);
+
 	if (result.path == NULL)
 		goto done;
 
@@ -1258,10 +1521,10 @@ static void properties_reply(DBusPendingCall *call, void *user_data)
 
 	frequency = (result.frequency < 0) ? 0 : result.frequency;
 
-	if (result.has_rsn == TRUE)
-		security = "rsn";
-	else if (result.has_wpa == TRUE)
-		security = "wpa";
+	if (result.has_8021x == TRUE)
+		security = "ieee8021x";
+	else if (result.has_psk == TRUE)
+		security = "psk";
 	else if (result.has_wep == TRUE)
 		security = "wep";
 	else
@@ -1273,6 +1536,13 @@ static void properties_reply(DBusPendingCall *call, void *user_data)
 					result.ssid, result.ssid_len,
 							mode, security);
 
+	if (result.has_psk == TRUE) {
+		if (result.has_rsn == TRUE)
+			security = "rsn";
+		else if (result.has_wpa == TRUE)
+			security = "wpa";
+	}
+
 	network = connman_device_get_network(task->device, result.path);
 	if (network == NULL) {
 		int index;
@@ -1327,8 +1597,11 @@ done:
 	g_free(result.name);
 	g_free(result.ssid);
 
+unref:
 	dbus_message_unref(reply);
 
+	dbus_pending_call_unref(call);
+
 	get_properties(task);
 }
 
@@ -1393,8 +1666,6 @@ static void scan_results_reply(DBusPendingCall *call, void *user_data)
 	DBG("task %p", task);
 
 	reply = dbus_pending_call_steal_reply(call);
-	if (reply == NULL)
-		goto noscan;
 
 	if (dbus_message_get_type(reply) == DBUS_MESSAGE_TYPE_ERROR)
 		goto done;
@@ -1428,6 +1699,8 @@ static void scan_results_reply(DBusPendingCall *call, void *user_data)
 
 	dbus_message_unref(reply);
 
+	dbus_pending_call_unref(call);
+
 	get_properties(task);
 
 	return;
@@ -1435,7 +1708,8 @@ static void scan_results_reply(DBusPendingCall *call, void *user_data)
 done:
 	dbus_message_unref(reply);
 
-noscan:
+	dbus_pending_call_unref(call);
+
 	task->result_call = NULL;
 
 	if (task->scanning == TRUE) {
@@ -1519,13 +1793,13 @@ static int task_connect(struct supplicant_task *task)
 
 	ssid = connman_network_get_blob(task->network, "WiFi.SSID", &ssid_len);
 
-	DBG("address %s security %s passphrase %s",
-					address, security, passphrase);
+	DBG("address %s security %s", address, security);
 
-	if (security == NULL && passphrase == NULL)
+	if (security == NULL)
 		return -EINVAL;
 
-	if (g_str_equal(security, "none") == FALSE && passphrase == NULL)
+	if (passphrase == NULL && g_str_equal(security, "none") == FALSE &&
+				g_str_equal(security, "ieee8021x") == FALSE)
 		return -EINVAL;
 
 	remove_network(task);
@@ -1534,7 +1808,9 @@ static int task_connect(struct supplicant_task *task)
 
 	add_network(task);
 
-	set_network(task, ssid, ssid_len, address, security, passphrase);
+	err = set_network(task, ssid, ssid_len, address, security, passphrase);
+	if (err < 0)
+		return err;
 
 	err = select_network(task);
 	if (err < 0)
@@ -1570,7 +1846,7 @@ static void state_change(struct supplicant_task *task, DBusMessage *msg)
 	const char *newstate, *oldstate;
 	unsigned char bssid[ETH_ALEN];
 	unsigned int bssid_len;
-	enum supplicant_state state;
+	enum supplicant_state state, prevstate;
 
 	dbus_error_init(&error);
 
@@ -1587,7 +1863,8 @@ static void state_change(struct supplicant_task *task, DBusMessage *msg)
 
 	DBG("state %s ==> %s", oldstate, newstate);
 
-	connman_info("%s %s", task->ifname, newstate);
+	connman_info("%s %s%s", task->ifname, newstate,
+				task->scanning == TRUE ? " (scanning)" : "");
 
 	state = string2state(newstate);
 	if (state == WPA_INVALID)
@@ -1598,6 +1875,7 @@ static void state_change(struct supplicant_task *task, DBusMessage *msg)
 		task->scanning = FALSE;
 	}
 
+	prevstate = task->state;
 	task->state = state;
 
 	if (task->network == NULL)
@@ -1605,6 +1883,17 @@ static void state_change(struct supplicant_task *task, DBusMessage *msg)
 
 	switch (task->state) {
 	case WPA_COMPLETED:
+		switch (prevstate) {
+		case WPA_ASSOCIATED:
+		case WPA_GROUP_HANDSHAKE:
+			break;
+		default:
+			goto badstate;
+		}
+
+		/* reset scan trigger and schedule background scan */
+		connman_device_schedule_scan(task->device);
+
 		if (get_bssid(task->device, bssid, &bssid_len) == 0)
 			connman_network_set_address(task->network,
 							bssid, bssid_len);
@@ -1613,30 +1902,30 @@ static void state_change(struct supplicant_task *task, DBusMessage *msg)
 		connman_network_set_connected(task->network, TRUE);
 		break;
 
-	case WPA_DISCONNECTED:
-		remove_network(task);
-
-		/* carrier off */
-		connman_network_set_connected(task->network, FALSE);
-
-		if (task->disconnecting == TRUE) {
-			connman_network_unref(task->network);
-			task->disconnecting = FALSE;
-
-			if (task->pending_network != NULL) {
-				task->network = task->pending_network;
-				task->pending_network = NULL;
-				task_connect(task);
-			} else
-				task->network = NULL;
+	case WPA_ASSOCIATING:
+		switch (prevstate) {
+		case WPA_COMPLETED:
+			break;
+		case WPA_SCANNING:
+			connman_network_set_associating(task->network, TRUE);
+			break;
+		default:
+			goto badstate;
 		}
 		break;
 
-	case WPA_ASSOCIATING:
-		connman_network_set_associating(task->network, TRUE);
-		break;
-
 	case WPA_INACTIVE:
+		switch (prevstate) {
+		case WPA_SCANNING:
+		case WPA_DISCONNECTED:
+			break;
+		default:
+			goto badstate;
+		}
+		/* fall through */
+
+	case WPA_DISCONNECTED:
+		/* carrier off */
 		connman_network_set_connected(task->network, FALSE);
 
 		if (task->disconnecting == TRUE) {
@@ -1656,29 +1945,31 @@ static void state_change(struct supplicant_task *task, DBusMessage *msg)
 		connman_network_set_associating(task->network, FALSE);
 		break;
 	}
+
+	return;
+
+badstate:
+	connman_error("%s invalid state change %s -> %s", task->ifname,
+							oldstate, newstate);
 }
 
-static DBusHandlerResult supplicant_filter(DBusConnection *conn,
+static gboolean supplicant_filter(DBusConnection *conn,
 						DBusMessage *msg, void *data)
 {
 	struct supplicant_task *task;
 	const char *member, *path;
 
-	if (dbus_message_has_interface(msg,
-				SUPPLICANT_INTF ".Interface") == FALSE)
-		return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
-
 	member = dbus_message_get_member(msg);
 	if (member == NULL)
-		return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
+		return TRUE;
 
 	path = dbus_message_get_path(msg);
 	if (path == NULL)
-		return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
+		return TRUE;
 
 	task = find_task_by_path(path);
 	if (task == NULL)
-		return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
+		return TRUE;
 
 	DBG("task %p member %s", task, member);
 
@@ -1689,7 +1980,7 @@ static DBusHandlerResult supplicant_filter(DBusConnection *conn,
 	else if (g_str_equal(member, "StateChange") == TRUE)
 		state_change(task, msg);
 
-	return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
+	return TRUE;
 }
 
 int supplicant_start(struct connman_device *device)
@@ -1711,8 +2002,8 @@ int supplicant_start(struct connman_device *device)
 		goto failed;
 	}
 
-	task->mac80211 = connman_inet_is_mac80211(task->ifindex);
-	if (task->mac80211 == FALSE)
+	task->cfg80211 = connman_inet_is_cfg80211(task->ifindex);
+	if (task->cfg80211 == FALSE)
 		connman_warn("Enabling quirks for unsupported driver");
 
 	task->range = g_try_malloc0(sizeof(struct iw_range));
@@ -1920,9 +2211,8 @@ static void supplicant_remove(DBusConnection *conn, void *user_data)
 	}
 }
 
-static const char *supplicant_rule = "type=signal,"
-				"interface=" SUPPLICANT_INTF ".Interface";
 static guint watch;
+static guint iface_watch;
 
 static int supplicant_create(void)
 {
@@ -1935,18 +2225,20 @@ static int supplicant_create(void)
 
 	DBG("connection %p", connection);
 
-	if (dbus_connection_add_filter(connection,
-				supplicant_filter, NULL, NULL) == FALSE) {
-		connection = connman_dbus_get_connection();
-		return -EIO;
-	}
-
-	dbus_bus_add_match(connection, supplicant_rule, NULL);
-	dbus_connection_flush(connection);
-
 	watch = g_dbus_add_service_watch(connection, SUPPLICANT_NAME,
 			supplicant_probe, supplicant_remove, NULL, NULL);
 
+	iface_watch = g_dbus_add_signal_watch(connection, NULL, NULL,
+						SUPPLICANT_INTF ".Interface",
+						NULL, supplicant_filter,
+						NULL, NULL);
+
+	if (watch == 0 || iface_watch == 0) {
+		g_dbus_remove_watch(connection, watch);
+		g_dbus_remove_watch(connection, iface_watch);
+		return -EIO;
+	}
+
 	return 0;
 }
 
@@ -1957,13 +2249,8 @@ static void supplicant_destroy(void)
 
 	DBG("connection %p", connection);
 
-	if (watch > 0)
-		g_dbus_remove_watch(connection, watch);
-
-	dbus_bus_remove_match(connection, supplicant_rule, NULL);
-	dbus_connection_flush(connection);
-
-	dbus_connection_remove_filter(connection, supplicant_filter, NULL);
+	g_dbus_remove_watch(connection, watch);
+	g_dbus_remove_watch(connection, iface_watch);
 
 	dbus_connection_unref(connection);
 	connection = NULL;
