@@ -2,7 +2,7 @@
  *
  *  Connection Manager
  *
- *  Copyright (C) 2007-2009  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2007-2010  Intel Corporation. All rights reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -24,6 +24,11 @@
 #endif
 
 #include <errno.h>
+#include <net/if.h>
+
+#ifndef IFF_LOWER_UP
+#define IFF_LOWER_UP	0x10000
+#endif
 
 #include <dbus/dbus.h>
 #include <glib.h>
@@ -31,6 +36,7 @@
 #define CONNMAN_API_SUBJECT_TO_CHANGE
 #include <connman/plugin.h>
 #include <connman/device.h>
+#include <connman/rtnl.h>
 #include <connman/log.h>
 
 #include "supplicant.h"
@@ -41,6 +47,9 @@
 struct wifi_data {
 	char *identifier;
 	connman_bool_t connected;
+	int index;
+	unsigned flags;
+	unsigned int watch;
 };
 
 static int network_probe(struct connman_network *network)
@@ -78,33 +87,70 @@ static struct connman_network_driver network_driver = {
 	.disconnect	= network_disconnect,
 };
 
+static void wifi_newlink(unsigned flags, unsigned change, void *user_data)
+{
+	struct connman_device *device = user_data;
+	struct wifi_data *wifi = connman_device_get_data(device);
+
+	DBG("index %d flags %d change %d", wifi->index, flags, change);
+
+	if ((wifi->flags & IFF_UP) != (flags & IFF_UP)) {
+		if (flags & IFF_UP) {
+			DBG("power on");
+			connman_device_set_powered(device, TRUE);
+			supplicant_scan(device);
+		} else {
+			connman_device_set_powered(device, FALSE);
+			DBG("power off");
+		}
+	}
+
+	if ((wifi->flags & IFF_LOWER_UP) != (flags & IFF_LOWER_UP)) {
+		if (flags & IFF_LOWER_UP) {
+			DBG("carrier on");
+		} else {
+			DBG("carrier off");
+		}
+	}
+
+	wifi->flags = flags;
+}
+
 static int wifi_probe(struct connman_device *device)
 {
-	struct wifi_data *data;
+	struct wifi_data *wifi;
 
 	DBG("device %p", device);
 
-	data = g_try_new0(struct wifi_data, 1);
-	if (data == NULL)
+	wifi = g_try_new0(struct wifi_data, 1);
+	if (wifi == NULL)
 		return -ENOMEM;
 
-	data->connected = FALSE;
+	wifi->connected = FALSE;
 
-	connman_device_set_data(device, data);
+	connman_device_set_data(device, wifi);
+
+	wifi->index = connman_device_get_index(device);
+	wifi->flags = 0;
+
+	wifi->watch = connman_rtnl_add_newlink_watch(wifi->index,
+							wifi_newlink, device);
 
 	return 0;
 }
 
 static void wifi_remove(struct connman_device *device)
 {
-	struct wifi_data *data = connman_device_get_data(device);
+	struct wifi_data *wifi = connman_device_get_data(device);
 
 	DBG("device %p", device);
 
 	connman_device_set_data(device, NULL);
 
-	g_free(data->identifier);
-	g_free(data);
+	connman_rtnl_remove_watch(wifi->watch);
+
+	g_free(wifi->identifier);
+	g_free(wifi);
 }
 
 static int wifi_enable(struct connman_device *device)
@@ -116,11 +162,11 @@ static int wifi_enable(struct connman_device *device)
 
 static int wifi_disable(struct connman_device *device)
 {
-	struct wifi_data *data = connman_device_get_data(device);
+	struct wifi_data *wifi = connman_device_get_data(device);
 
 	DBG("device %p", device);
 
-	data->connected = FALSE;
+	wifi->connected = FALSE;
 
 	return supplicant_stop(device);
 }
@@ -132,26 +178,6 @@ static int wifi_scan(struct connman_device *device)
 	return supplicant_scan(device);
 }
 
-static int wifi_join(struct connman_device *device,
-					struct connman_network *network)
-{
-	int err;
-
-	DBG("device %p", device);
-
-	err = supplicant_connect(network);
-	if (err < 0)
-		return err;
-
-	connman_network_ref(network);
-
-	connman_device_add_network(device, network);
-
-	connman_network_set_available(network, TRUE);
-
-	return 0;
-}
-
 static struct connman_device_driver wifi_driver = {
 	.name		= "wifi",
 	.type		= CONNMAN_DEVICE_TYPE_WIFI,
@@ -160,7 +186,6 @@ static struct connman_device_driver wifi_driver = {
 	.enable		= wifi_enable,
 	.disable	= wifi_disable,
 	.scan		= wifi_scan,
-	.join		= wifi_join,
 };
 
 static void wifi_register(void)
