@@ -2,7 +2,7 @@
  *
  *  Connection Manager
  *
- *  Copyright (C) 2007-2009  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2007-2010  Intel Corporation. All rights reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -23,13 +23,14 @@
 #include <config.h>
 #endif
 
+#include <stdio.h>
 #include <unistd.h>
 #include <string.h>
 #include <sys/socket.h>
 #include <arpa/inet.h>
 #include <netinet/ether.h>
-#include <net/if.h>
 #include <net/if_arp.h>
+#include <linux/if.h>
 #include <linux/netlink.h>
 #include <linux/rtnetlink.h>
 
@@ -43,6 +44,7 @@
 struct watch_data {
 	unsigned int id;
 	int index;
+	connman_rtnl_operstate_cb_t operstate;
 	connman_rtnl_link_cb_t newlink;
 	void *user_data;
 };
@@ -50,6 +52,45 @@ struct watch_data {
 static GSList *watch_list = NULL;
 static unsigned int watch_id = 0;
 
+/**
+ * connman_rtnl_add_operstate_watch:
+ * @index: network device index
+ * @callback: callback function
+ * @user_data: callback data;
+ *
+ * Add a new RTNL watch for operation state events
+ *
+ * Returns: %0 on failure and a unique id on success
+ */
+unsigned int connman_rtnl_add_operstate_watch(int index,
+			connman_rtnl_operstate_cb_t callback, void *user_data)
+{
+	struct watch_data *watch;
+
+	watch = g_try_new0(struct watch_data, 1);
+	if (watch == NULL)
+		return 0;
+
+	watch->id = ++watch_id;
+	watch->index = index;
+
+	watch->operstate = callback;
+	watch->user_data = user_data;
+
+	watch_list = g_slist_prepend(watch_list, watch);
+
+	DBG("id %d", watch->id);
+
+	if (callback) {
+		unsigned char operstate = 0;
+
+		if (operstate > 0)
+			callback(operstate, user_data);
+	}
+
+	return watch->id;
+}
+
 /**
  * connman_rtnl_add_newlink_watch:
  * @index: network device index
@@ -177,19 +218,106 @@ void connman_rtnl_unregister(struct connman_rtnl *rtnl)
 	rtnl_list = g_slist_remove(rtnl_list, rtnl);
 }
 
-static void process_newlink(unsigned short type, int index,
-					unsigned flags, unsigned change)
+static const char *operstate2str(unsigned char operstate)
 {
+	switch (operstate) {
+	case IF_OPER_UNKNOWN:
+		return "UNKNOWN";
+	case IF_OPER_NOTPRESENT:
+		return "NOT-PRESENT";
+	case IF_OPER_DOWN:
+		return "DOWN";
+	case IF_OPER_LOWERLAYERDOWN:
+		return "LOWER-LAYER-DOWN";
+	case IF_OPER_TESTING:
+		return "TESTING";
+	case IF_OPER_DORMANT:
+		return "DORMANT";
+	case IF_OPER_UP:
+		return "UP";
+	}
+
+	return "";
+}
+
+static void extract_link(struct ifinfomsg *msg, int bytes,
+				struct ether_addr *address, const char **ifname,
+				unsigned int *mtu, unsigned char *operstate,
+						struct rtnl_link_stats *stats)
+{
+	struct rtattr *attr;
+
+	for (attr = IFLA_RTA(msg); RTA_OK(attr, bytes);
+					attr = RTA_NEXT(attr, bytes)) {
+		switch (attr->rta_type) {
+		case IFLA_ADDRESS:
+			if (address != NULL)
+				memcpy(address, RTA_DATA(attr), ETH_ALEN);
+			break;
+		case IFLA_IFNAME:
+			if (ifname != NULL)
+				*ifname = RTA_DATA(attr);
+			break;
+		case IFLA_MTU:
+			if (mtu != NULL)
+				*mtu = *((unsigned int *) RTA_DATA(attr));
+			break;
+		case IFLA_STATS:
+			if (stats != NULL)
+				memcpy(stats, RTA_DATA(attr),
+					sizeof(struct rtnl_link_stats));
+			break;
+		case IFLA_OPERSTATE:
+			if (operstate != NULL)
+				*operstate = *((unsigned char *) RTA_DATA(attr));
+			break;
+		case IFLA_LINKMODE:
+			break;
+		}
+	}
+}
+
+static void process_newlink(unsigned short type, int index, unsigned flags,
+			unsigned change, struct ifinfomsg *msg, int bytes)
+{
+	struct ether_addr address = {{ 0, 0, 0, 0, 0, 0 }};
+	struct ether_addr compare = {{ 0, 0, 0, 0, 0, 0 }};
+	struct rtnl_link_stats stats;
+	unsigned char operstate = 0xff;
+	const char *ifname = NULL;
+	unsigned int mtu = 0;
+	char str[18];
 	GSList *list;
 
+	memset(&stats, 0, sizeof(stats));
+	extract_link(msg, bytes, &address, &ifname, &mtu, &operstate, &stats);
+
+	snprintf(str, 18, "%02X:%02X:%02X:%02X:%02X:%02X",
+						address.ether_addr_octet[0],
+						address.ether_addr_octet[1],
+						address.ether_addr_octet[2],
+						address.ether_addr_octet[3],
+						address.ether_addr_octet[4],
+						address.ether_addr_octet[5]);
+
 	switch (type) {
 	case ARPHRD_ETHER:
 	case ARPHRD_LOOPBACK:
 	case ARPHRD_NONE:
-		__connman_ipconfig_newlink(index, type, flags);
+		__connman_ipconfig_newlink(index, type, flags,
+							str, mtu, &stats);
 		break;
 	}
 
+	if (memcmp(&address, &compare, ETH_ALEN) != 0)
+		connman_info("%s {newlink} index %d address %s mtu %u",
+						ifname, index, str, mtu);
+
+	if (operstate != 0xff)
+		connman_info("%s {newlink} index %d operstate %u <%s>",
+						ifname, index, operstate,
+						operstate2str(operstate));
+
 	for (list = rtnl_list; list; list = list->next) {
 		struct connman_rtnl *rtnl = list->data;
 
@@ -203,16 +331,40 @@ static void process_newlink(unsigned short type, int index,
 		if (watch->index != index)
 			continue;
 
+		if (operstate != 0xff && watch->operstate)
+			watch->operstate(operstate, watch->user_data);
+
 		if (watch->newlink)
 			watch->newlink(flags, change, watch->user_data);
 	}
 }
 
-static void process_dellink(unsigned short type, int index,
-					unsigned flags, unsigned change)
+static void process_dellink(unsigned short type, int index, unsigned flags,
+			unsigned change, struct ifinfomsg *msg, int bytes)
 {
+	struct rtnl_link_stats stats;
+	unsigned char operstate = 0xff;
+	const char *ifname = NULL;
 	GSList *list;
 
+	memset(&stats, 0, sizeof(stats));
+	extract_link(msg, bytes, NULL, &ifname, NULL, &operstate, &stats);
+
+	if (operstate != 0xff)
+		connman_info("%s {dellink} index %d operstate %u <%s>",
+						ifname, index, operstate,
+						operstate2str(operstate));
+
+	for (list = watch_list; list; list = list->next) {
+		struct watch_data *watch = list->data;
+
+		if (watch->index != index)
+			continue;
+
+		if (operstate != 0xff && watch->operstate)
+			watch->operstate(operstate, watch->user_data);
+	}
+
 	for (list = rtnl_list; list; list = list->next) {
 		struct connman_rtnl *rtnl = list->data;
 
@@ -224,7 +376,7 @@ static void process_dellink(unsigned short type, int index,
 	case ARPHRD_ETHER:
 	case ARPHRD_LOOPBACK:
 	case ARPHRD_NONE:
-		__connman_ipconfig_dellink(index);
+		__connman_ipconfig_dellink(index, &stats);
 		break;
 	}
 }
@@ -263,8 +415,8 @@ static void extract_addr(struct ifaddrmsg *msg, int bytes,
 static void process_newaddr(unsigned char family, unsigned char prefixlen,
 				int index, struct ifaddrmsg *msg, int bytes)
 {
-	const char *label;
 	struct in_addr address = { INADDR_ANY };
+	const char *label = NULL;
 
 	if (family != AF_INET)
 		return;
@@ -278,8 +430,8 @@ static void process_newaddr(unsigned char family, unsigned char prefixlen,
 static void process_deladdr(unsigned char family, unsigned char prefixlen,
 				int index, struct ifaddrmsg *msg, int bytes)
 {
-	const char *label;
 	struct in_addr address = { INADDR_ANY };
+	const char *label = NULL;
 
 	if (family != AF_INET)
 		return;
@@ -504,8 +656,8 @@ static void rtnl_newlink(struct nlmsghdr *hdr)
 
 	rtnl_link(hdr);
 
-	process_newlink(msg->ifi_type, msg->ifi_index,
-					msg->ifi_flags, msg->ifi_change);
+	process_newlink(msg->ifi_type, msg->ifi_index, msg->ifi_flags,
+				msg->ifi_change, msg, IFA_PAYLOAD(hdr));
 }
 
 static void rtnl_dellink(struct nlmsghdr *hdr)
@@ -514,8 +666,8 @@ static void rtnl_dellink(struct nlmsghdr *hdr)
 
 	rtnl_link(hdr);
 
-	process_dellink(msg->ifi_type, msg->ifi_index,
-					msg->ifi_flags, msg->ifi_change);
+	process_dellink(msg->ifi_type, msg->ifi_index, msg->ifi_flags,
+				msg->ifi_change, msg, IFA_PAYLOAD(hdr));
 }
 
 static void rtnl_addr(struct nlmsghdr *hdr)
@@ -899,6 +1051,11 @@ static int send_getroute(void)
 	return queue_request(req);
 }
 
+int __connman_rtnl_request_update(void)
+{
+	return send_getlink();
+}
+
 int __connman_rtnl_init(void)
 {
 	struct sockaddr_nl addr;
