@@ -2,7 +2,7 @@
  *
  *  Connection Manager
  *
- *  Copyright (C) 2007-2009  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2007-2010  Intel Corporation. All rights reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -75,6 +75,7 @@ struct connman_service {
 	char *phase2;
 	DBusMessage *pending;
 	guint timeout;
+	struct connman_location *location;
 };
 
 static void append_path(gpointer value, gpointer user_data)
@@ -139,6 +140,8 @@ const char *__connman_service_type2string(enum connman_service_type type)
 		return "bluetooth";
 	case CONNMAN_SERVICE_TYPE_CELLULAR:
 		return "cellular";
+	case CONNMAN_SERVICE_TYPE_GPS:
+		return "gps";
 	case CONNMAN_SERVICE_TYPE_VPN:
 		return "vpn";
 	}
@@ -201,6 +204,10 @@ static const char *state2string(enum connman_service_state state)
 		return "configuration";
 	case CONNMAN_SERVICE_STATE_READY:
 		return "ready";
+	case CONNMAN_SERVICE_STATE_LOGIN:
+		return "login";
+	case CONNMAN_SERVICE_STATE_ONLINE:
+		return "online";
 	case CONNMAN_SERVICE_STATE_DISCONNECT:
 		return "disconnect";
 	case CONNMAN_SERVICE_STATE_FAILURE:
@@ -312,13 +319,29 @@ static void strength_changed(struct connman_service *service)
 
 static void favorite_changed(struct connman_service *service)
 {
+	if (service->path == NULL)
+		return;
+
 	connman_dbus_property_changed_basic(service->path,
 				CONNMAN_SERVICE_INTERFACE, "Favorite",
 					DBUS_TYPE_BOOLEAN, &service->favorite);
 }
 
+static void immutable_changed(struct connman_service *service)
+{
+	if (service->path == NULL)
+		return;
+
+	connman_dbus_property_changed_basic(service->path,
+				CONNMAN_SERVICE_INTERFACE, "Immutable",
+					DBUS_TYPE_BOOLEAN, &service->immutable);
+}
+
 static void roaming_changed(struct connman_service *service)
 {
+	if (service->path == NULL)
+		return;
+
 	connman_dbus_property_changed_basic(service->path,
 				CONNMAN_SERVICE_INTERFACE, "Roaming",
 					DBUS_TYPE_BOOLEAN, &service->roaming);
@@ -326,6 +349,9 @@ static void roaming_changed(struct connman_service *service)
 
 static void autoconnect_changed(struct connman_service *service)
 {
+	if (service->path == NULL)
+		return;
+
 	connman_dbus_property_changed_basic(service->path,
 				CONNMAN_SERVICE_INTERFACE, "AutoConnect",
 				DBUS_TYPE_BOOLEAN, &service->autoconnect);
@@ -342,6 +368,7 @@ static void passphrase_changed(struct connman_service *service)
 	case CONNMAN_SERVICE_TYPE_WIMAX:
 	case CONNMAN_SERVICE_TYPE_BLUETOOTH:
 	case CONNMAN_SERVICE_TYPE_CELLULAR:
+	case CONNMAN_SERVICE_TYPE_GPS:
 	case CONNMAN_SERVICE_TYPE_VPN:
 		return;
 	case CONNMAN_SERVICE_TYPE_WIFI:
@@ -355,10 +382,11 @@ static void passphrase_changed(struct connman_service *service)
 		case CONNMAN_SERVICE_SECURITY_PSK:
 		case CONNMAN_SERVICE_SECURITY_WPA:
 		case CONNMAN_SERVICE_SECURITY_RSN:
-		case CONNMAN_SERVICE_SECURITY_8021X:
 			if (service->passphrase == NULL)
 				required = TRUE;
 			break;
+		case CONNMAN_SERVICE_SECURITY_8021X:
+			break;
 		}
 		break;
 	}
@@ -379,6 +407,7 @@ static void apn_changed(struct connman_service *service)
 	case CONNMAN_SERVICE_TYPE_WIMAX:
 	case CONNMAN_SERVICE_TYPE_BLUETOOTH:
 	case CONNMAN_SERVICE_TYPE_WIFI:
+	case CONNMAN_SERVICE_TYPE_GPS:
 	case CONNMAN_SERVICE_TYPE_VPN:
 		return;
 	case CONNMAN_SERVICE_TYPE_CELLULAR:
@@ -405,6 +434,8 @@ static void append_ethernet(DBusMessageIter *iter, void *user_data)
 	case CONNMAN_SERVICE_STATE_ASSOCIATION:
 	case CONNMAN_SERVICE_STATE_CONFIGURATION:
 	case CONNMAN_SERVICE_STATE_READY:
+	case CONNMAN_SERVICE_STATE_LOGIN:
+	case CONNMAN_SERVICE_STATE_ONLINE:
 		break;
 	}
 
@@ -425,6 +456,8 @@ static void append_ipv4(DBusMessageIter *iter, void *user_data)
 	case CONNMAN_SERVICE_STATE_CONFIGURATION:
 		return;
 	case CONNMAN_SERVICE_STATE_READY:
+	case CONNMAN_SERVICE_STATE_LOGIN:
+	case CONNMAN_SERVICE_STATE_ONLINE:
 		break;
 	}
 
@@ -440,6 +473,28 @@ static void append_ipv4config(DBusMessageIter *iter, void *user_data)
 		__connman_ipconfig_append_ipv4config(service->ipconfig, iter);
 }
 
+static void append_proxy(DBusMessageIter *iter, void *user_data)
+{
+	struct connman_service *service = user_data;
+
+	switch (service->state) {
+	case CONNMAN_SERVICE_STATE_UNKNOWN:
+	case CONNMAN_SERVICE_STATE_IDLE:
+	case CONNMAN_SERVICE_STATE_FAILURE:
+	case CONNMAN_SERVICE_STATE_DISCONNECT:
+	case CONNMAN_SERVICE_STATE_ASSOCIATION:
+	case CONNMAN_SERVICE_STATE_CONFIGURATION:
+		return;
+	case CONNMAN_SERVICE_STATE_READY:
+	case CONNMAN_SERVICE_STATE_LOGIN:
+	case CONNMAN_SERVICE_STATE_ONLINE:
+		break;
+	}
+
+	if (service->ipconfig != NULL)
+		__connman_ipconfig_append_proxy(service->ipconfig, iter);
+}
+
 static void settings_changed(struct connman_service *service)
 {
 	connman_dbus_property_changed_dict(service->path,
@@ -447,6 +502,15 @@ static void settings_changed(struct connman_service *service)
 							append_ipv4, service);
 }
 
+static void ipv4_configuration_changed(struct connman_service *service)
+{
+	connman_dbus_property_changed_dict(service->path,
+					CONNMAN_SERVICE_INTERFACE,
+							"IPv4.Configuration",
+							append_ipv4config,
+							service);
+}
+
 static DBusMessage *get_properties(DBusConnection *conn,
 					DBusMessage *msg, void *user_data)
 {
@@ -515,8 +579,7 @@ static DBusMessage *get_properties(DBusConnection *conn,
 	switch (service->type) {
 	case CONNMAN_SERVICE_TYPE_UNKNOWN:
 	case CONNMAN_SERVICE_TYPE_SYSTEM:
-	case CONNMAN_SERVICE_TYPE_WIMAX:
-	case CONNMAN_SERVICE_TYPE_BLUETOOTH:
+	case CONNMAN_SERVICE_TYPE_GPS:
 	case CONNMAN_SERVICE_TYPE_VPN:
 		break;
 	case CONNMAN_SERVICE_TYPE_CELLULAR:
@@ -568,16 +631,19 @@ static DBusMessage *get_properties(DBusConnection *conn,
 		case CONNMAN_SERVICE_SECURITY_PSK:
 		case CONNMAN_SERVICE_SECURITY_WPA:
 		case CONNMAN_SERVICE_SECURITY_RSN:
-		case CONNMAN_SERVICE_SECURITY_8021X:
 			if (service->passphrase == NULL)
 				required = TRUE;
 			break;
+		case CONNMAN_SERVICE_SECURITY_8021X:
+			break;
 		}
 
 		connman_dbus_dict_append_basic(&dict, "PassphraseRequired",
 						DBUS_TYPE_BOOLEAN, &required);
 		/* fall through */
 	case CONNMAN_SERVICE_TYPE_ETHERNET:
+	case CONNMAN_SERVICE_TYPE_WIMAX:
+	case CONNMAN_SERVICE_TYPE_BLUETOOTH:
 		connman_dbus_dict_append_dict(&dict, "Ethernet",
 						append_ethernet, service);
 		break;
@@ -588,6 +654,8 @@ static DBusMessage *get_properties(DBusConnection *conn,
 	connman_dbus_dict_append_dict(&dict, "IPv4.Configuration",
 						append_ipv4config, service);
 
+	connman_dbus_dict_append_dict(&dict, "Proxy", append_proxy, service);
+
 	connman_dbus_dict_close(&array, &dict);
 
 	return reply;
@@ -727,6 +795,8 @@ static DBusMessage *set_property(DBusConnection *conn,
 		if (err < 0)
 			return __connman_error_failed(msg, -err);
 
+		ipv4_configuration_changed(service);
+
 		__connman_storage_save_service(service);
 	} else
 		return __connman_error_invalid_property(msg);
@@ -782,6 +852,8 @@ static connman_bool_t is_connecting(struct connman_service *service)
 	case CONNMAN_SERVICE_STATE_FAILURE:
 	case CONNMAN_SERVICE_STATE_DISCONNECT:
 	case CONNMAN_SERVICE_STATE_READY:
+	case CONNMAN_SERVICE_STATE_LOGIN:
+	case CONNMAN_SERVICE_STATE_ONLINE:
 		break;
 	case CONNMAN_SERVICE_STATE_ASSOCIATION:
 	case CONNMAN_SERVICE_STATE_CONFIGURATION:
@@ -990,9 +1062,6 @@ static DBusMessage *connect_service(DBusConnection *conn,
 		return NULL;
 	}
 
-	dbus_message_unref(service->pending);
-	service->pending = NULL;
-
 	return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
 }
 
@@ -1059,14 +1128,15 @@ static DBusMessage *remove_service(DBusConnection *conn,
 
 	set_idle(service);
 
-	connman_service_set_favorite(service, FALSE);
+	__connman_service_set_favorite(service, FALSE);
 	__connman_storage_save_service(service);
 
 	return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
 }
 
-static DBusMessage *move_before(DBusConnection *conn,
-					DBusMessage *msg, void *user_data)
+static DBusMessage *move_service(DBusConnection *conn,
+					DBusMessage *msg, void *user_data,
+								gboolean before)
 {
 	struct connman_service *service = user_data;
 	struct connman_service *target;
@@ -1096,41 +1166,23 @@ static DBusMessage *move_before(DBusConnection *conn,
 	src = g_hash_table_lookup(service_hash, service->identifier);
 	dst = g_hash_table_lookup(service_hash, target->identifier);
 
-	g_sequence_move(src, dst);
+	before ? g_sequence_move(src, dst) : g_sequence_move(dst, src);
 
 	__connman_profile_changed(FALSE);
 
 	return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
 }
 
+static DBusMessage *move_before(DBusConnection *conn,
+					DBusMessage *msg, void *user_data)
+{
+	return move_service(conn, msg, user_data, TRUE);
+}
+
 static DBusMessage *move_after(DBusConnection *conn,
 					DBusMessage *msg, void *user_data)
 {
-	struct connman_service *service = user_data;
-	struct connman_service *target;
-	const char *path;
-
-	DBG("service %p", service);
-
-	dbus_message_get_args(msg, NULL, DBUS_TYPE_OBJECT_PATH, &path,
-							DBUS_TYPE_INVALID);
-
-	if (service->favorite == FALSE)
-		return __connman_error_not_supported(msg);
-
-	target = find_service(path);
-	if (target == NULL || target->favorite == FALSE || target == service)
-		return __connman_error_invalid_service(msg);
-
-	DBG("target %s", target->identifier);
-
-	if (target->state != service->state)
-		return __connman_error_invalid_service(msg);
-
-	g_get_current_time(&service->modified);
-	__connman_storage_save_service(service);
-
-	return __connman_error_not_implemented(msg);
+	return move_service(conn, msg, user_data, FALSE);
 }
 
 static GDBusMethodTable service_methods[] = {
@@ -1180,6 +1232,9 @@ static void service_free(gpointer user_data)
 		service->ipconfig = NULL;
 	}
 
+	if (service->location != NULL)
+		connman_location_unref(service->location);
+
 	g_free(service->mcc);
 	g_free(service->mnc);
 	g_free(service->apn);
@@ -1265,9 +1320,16 @@ struct connman_service *connman_service_create(void)
 
 	__connman_service_initialize(service);
 
+	service->location = __connman_location_create(service);
+
 	return service;
 }
 
+struct connman_location *__connman_service_get_location(struct connman_service *service)
+{
+	return service->location;
+}
+
 /**
  * connman_service_ref:
  * @service: service structure
@@ -1322,6 +1384,7 @@ static gint service_compare(gconstpointer a, gconstpointer b,
 		case CONNMAN_SERVICE_TYPE_UNKNOWN:
 		case CONNMAN_SERVICE_TYPE_SYSTEM:
 		case CONNMAN_SERVICE_TYPE_ETHERNET:
+		case CONNMAN_SERVICE_TYPE_GPS:
 		case CONNMAN_SERVICE_TYPE_VPN:
 			break;
 		case CONNMAN_SERVICE_TYPE_WIFI:
@@ -1386,14 +1449,23 @@ __connman_service_get_network(struct connman_service *service)
 	return service->network;
 }
 
+struct connman_ipconfig *
+__connman_service_get_ipconfig(struct connman_service *service)
+{
+	if (service == NULL)
+		return NULL;
+
+	return service->ipconfig;
+}
+
 /**
- * connman_service_set_favorite:
+ * __connman_service_set_favorite:
  * @service: service structure
  * @favorite: favorite value
  *
  * Change the favorite setting of service
  */
-int connman_service_set_favorite(struct connman_service *service,
+int __connman_service_set_favorite(struct connman_service *service,
 						connman_bool_t favorite)
 {
 	GSequenceIter *iter;
@@ -1416,6 +1488,16 @@ int connman_service_set_favorite(struct connman_service *service,
 	return 0;
 }
 
+int __connman_service_set_immutable(struct connman_service *service,
+						connman_bool_t immutable)
+{
+	service->immutable = immutable;
+
+	immutable_changed(service);
+
+	return 0;
+}
+
 void __connman_service_set_string(struct connman_service *service,
 				  const char *key, const char *value)
 {
@@ -1477,6 +1559,9 @@ int __connman_service_indicate_state(struct connman_service *service,
 		__connman_service_disconnect(service);
 	}
 
+	if (state == CONNMAN_SERVICE_STATE_CONFIGURATION)
+		__connman_ipconfig_enable(service->ipconfig);
+
 	service->state = state;
 	state_changed(service);
 
@@ -1491,7 +1576,7 @@ int __connman_service_indicate_state(struct connman_service *service,
 	if (state == CONNMAN_SERVICE_STATE_READY) {
 		set_reconnect_state(service, TRUE);
 
-		connman_service_set_favorite(service, TRUE);
+		__connman_service_set_favorite(service, TRUE);
 
 		reply_pending(service, 0);
 
@@ -1503,7 +1588,11 @@ int __connman_service_indicate_state(struct connman_service *service,
 		__connman_notifier_connect(service->type);
 
 		default_changed();
+
+		__connman_location_detect(service);
 	} else if (state == CONNMAN_SERVICE_STATE_DISCONNECT) {
+		__connman_location_finish(service);
+
 		default_changed();
 
 		__connman_notifier_disconnect(service->type);
@@ -1579,8 +1668,8 @@ static connman_bool_t prepare_network(struct connman_service *service)
 	case CONNMAN_NETWORK_TYPE_WIMAX:
 	case CONNMAN_NETWORK_TYPE_BLUETOOTH_PAN:
 	case CONNMAN_NETWORK_TYPE_BLUETOOTH_DUN:
-	case CONNMAN_NETWORK_TYPE_CELLULAR:
 		break;
+	case CONNMAN_NETWORK_TYPE_CELLULAR:
 	case CONNMAN_NETWORK_TYPE_MBM:
 	case CONNMAN_NETWORK_TYPE_HSO:
 		connman_network_set_string(service->network,
@@ -1598,34 +1687,34 @@ static connman_bool_t prepare_network(struct connman_service *service)
 
 static void prepare_8021x(struct connman_service *service)
 {
-	if (service->eap)
+	if (service->eap != NULL)
 		connman_network_set_string(service->network, "WiFi.EAP",
 								service->eap);
 
-	if (service->identity)
+	if (service->identity != NULL)
 		connman_network_set_string(service->network, "WiFi.Identity",
 							service->identity);
 
-	if (service->ca_cert_file)
+	if (service->ca_cert_file != NULL)
 		connman_network_set_string(service->network, "WiFi.CACertFile",
 							service->ca_cert_file);
 
-	if (service->client_cert_file)
+	if (service->client_cert_file != NULL)
 		connman_network_set_string(service->network,
 						"WiFi.ClientCertFile",
 						service->client_cert_file);
 
-	if (service->private_key_file)
+	if (service->private_key_file != NULL)
 		connman_network_set_string(service->network,
 						"WiFi.PrivateKeyFile",
 						service->private_key_file);
 
-	if (service->private_key_passphrase)
+	if (service->private_key_passphrase != NULL)
 		connman_network_set_string(service->network,
 					"WiFi.PrivateKeyPassphrase",
 					service->private_key_passphrase);
 
-	if (service->phase2)
+	if (service->phase2 != NULL)
 		connman_network_set_string(service->network, "WiFi.Phase2",
 							service->phase2);
 }
@@ -1645,6 +1734,7 @@ int __connman_service_connect(struct connman_service *service)
 	switch (service->type) {
 	case CONNMAN_SERVICE_TYPE_UNKNOWN:
 	case CONNMAN_SERVICE_TYPE_SYSTEM:
+	case CONNMAN_SERVICE_TYPE_GPS:
 	case CONNMAN_SERVICE_TYPE_VPN:
 		return -EINVAL;
 	case CONNMAN_SERVICE_TYPE_ETHERNET:
@@ -1664,10 +1754,11 @@ int __connman_service_connect(struct connman_service *service)
 		case CONNMAN_SERVICE_SECURITY_PSK:
 		case CONNMAN_SERVICE_SECURITY_WPA:
 		case CONNMAN_SERVICE_SECURITY_RSN:
-		case CONNMAN_SERVICE_SECURITY_8021X:
 			if (service->passphrase == NULL)
 				return -ENOKEY;
 			break;
+		case CONNMAN_SERVICE_SECURITY_8021X:
+			break;
 		}
 		break;
 	}
@@ -1721,6 +1812,8 @@ int __connman_service_disconnect(struct connman_service *service)
 	} else
 		return -EOPNOTSUPP;
 
+	__connman_ipconfig_clear_address(service->ipconfig);
+
 	__connman_ipconfig_disable(service->ipconfig);
 
 	if (err < 0) {
@@ -1938,9 +2031,8 @@ failed:
 		if (network != NULL) {
 			connman_network_set_available(network, FALSE);
 			__connman_device_cleanup_networks(device);
-		}
-
-		__connman_service_put(service);
+		} else
+			__connman_service_put(service);
 	}
 
 	return err;
@@ -2079,6 +2171,15 @@ static void setup_ipconfig(struct connman_service *service, int index)
 	connman_ipconfig_set_ops(service->ipconfig, &service_ops);
 }
 
+void __connman_service_create_ipconfig(struct connman_service *service,
+								int index)
+{
+	if (service->ipconfig != NULL)
+		return;
+
+	setup_ipconfig(service, index);
+}
+
 /**
  * __connman_service_lookup_from_network:
  * @network: network structure
@@ -2334,6 +2435,7 @@ struct connman_service *__connman_service_create_from_network(struct connman_net
 	case CONNMAN_SERVICE_TYPE_ETHERNET:
 	case CONNMAN_SERVICE_TYPE_WIMAX:
 	case CONNMAN_SERVICE_TYPE_BLUETOOTH:
+	case CONNMAN_SERVICE_TYPE_GPS:
 	case CONNMAN_SERVICE_TYPE_VPN:
 		service->autoconnect = FALSE;
 		break;
@@ -2465,6 +2567,7 @@ static int service_load(struct connman_service *service)
 	case CONNMAN_SERVICE_TYPE_UNKNOWN:
 	case CONNMAN_SERVICE_TYPE_SYSTEM:
 	case CONNMAN_SERVICE_TYPE_ETHERNET:
+	case CONNMAN_SERVICE_TYPE_GPS:
 	case CONNMAN_SERVICE_TYPE_VPN:
 		break;
 	case CONNMAN_SERVICE_TYPE_WIFI:
@@ -2611,6 +2714,7 @@ update:
 	case CONNMAN_SERVICE_TYPE_UNKNOWN:
 	case CONNMAN_SERVICE_TYPE_SYSTEM:
 	case CONNMAN_SERVICE_TYPE_ETHERNET:
+	case CONNMAN_SERVICE_TYPE_GPS:
 	case CONNMAN_SERVICE_TYPE_VPN:
 		break;
 	case CONNMAN_SERVICE_TYPE_WIFI:
